import asyncio
import json
import logging
import re
import secrets
from ipaddress import ip_address, ip_network
from operator import attrgetter
from time import time
from typing import Any, AsyncGenerator, Awaitable, Callable, Dict, List, Literal, Optional, Set

from sentry_sdk import capture_event
from sentry_sdk.utils import event_from_exception, exc_info_from_error
from starlette.applications import Starlette
from starlette.datastructures import URL
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from starlette.responses import RedirectResponse, Response
from starlette.routing import get_name as get_endpoint_name

from . import glove
from .exceptions import UnexpectedResponse
from .utils import get_ip

logger = logging.getLogger('foxglove.middleware')
request_logger = logging.getLogger('foxglove.bad_requests')
CallNext = Callable[[Request], Awaitable[Response]]

__all__ = (
    'ErrorMiddleware',
    'CsrfMiddleware',
    'HostRedirectMiddleware',
    'CloudflareCheckMiddleware',
    'request_log_extra',
    'get_session_id',
    'update_session_id',
)


class ErrorMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: Starlette,
        should_warn: Callable[[Response], bool] = None,
        get_user: Callable[[Request], Awaitable[Dict[str, Any]]] = None,
    ):
        super().__init__(app)
        self.custom_should_warn = should_warn
        self.get_user = get_user

        from .main import glove

        self.glove = glove

    async def dispatch(self, request: Request, call_next: CallNext) -> Response:
        try:
            request.state.start_time = get_request_start(request)

            try:
                response = await call_next(request)
            except Exception as exc:
                await self.log(request, exc=exc)
                return Response('Internal Server Error', media_type='text/plain', status_code=500)
            else:
                if self.should_warn(response):
                    await self.log(request, response=response)
                return response
        except Exception:  # pragma: no cover
            # not sure if this is required, but better to keep it
            logger.critical('unhandled error in ErrorMiddleware', exc_info=True)
            raise

    async def log(
        self, request: Request, *, exc: Optional[Exception] = None, response: Optional[Response] = None
    ) -> None:
        event_data = await request_log_extra(request, exc, response)
        event_data['user'] = await self.user_info(request)
        view_ref = event_data['transaction']

        if exc:
            level = 'error'
            message = f'"{line_one(request)}", {exc!r}'
            fingerprint = view_ref, request.method, repr(exc)
            request_logger.exception(message, extra=event_data)
        else:
            assert response is not None
            level = 'warning'
            message = f'"{line_one(request)}", unexpected response: {response.status_code}'
            request_logger.warning(message, extra=event_data)
            fingerprint = view_ref, request.method, str(response.status_code)

        if glove.settings.sentry_dsn:
            hint = None
            if exc:
                exc_data, hint = event_from_exception(exc_info_from_error(exc))
                event_data.update(exc_data)

            event_data.update(message=message, level=level, logger='foxglove.request_errors', fingerprint=fingerprint)
            if not capture_event(event_data, hint):
                logger.critical(
                    'sentry not configured correctly, not sending message: %s',
                    message,
                    extra={'event_data': event_data},
                )

    def should_warn(self, response: Response) -> bool:
        if self.custom_should_warn:
            return self.custom_should_warn(response)
        else:
            return response.status_code > 310

    async def user_info(self, request: Request) -> Dict[str, Any]:
        user = dict(ip_address=get_ip(request))
        if get_user := self.get_user:
            try:
                user.update(await get_user(request))
            except Exception:
                logger.exception('error getting user for middleware logging')
        return user

    @staticmethod
    async def response_body(response: Response) -> bytes:
        if hasattr(response, 'body'):
            return response.body
        else:
            body_chunks = []
            async for chunk in response.body_iterator:
                if not isinstance(chunk, bytes):
                    chunk = chunk.encode(response.charset)
                body_chunks.append(chunk)

            response.body_iterator = async_gen_list(body_chunks)
            return b''.join(body_chunks)


async def request_log_extra(
    request: Request, exc: Optional[Exception] = None, response: Optional[Response] = None
) -> Dict[str, Any]:
    extra = dict(query=dict(request.query_params))

    if start_time := getattr(request.state, 'start_time', None):
        extra['duration'] = f'{(time() - start_time) * 1000:0.2f}ms'

    if endpoint := request.scope.get('endpoint'):
        extra.update(route_endpoint=get_endpoint_name(endpoint), path_params=dict(request.path_params))

    if exc:
        extra['exception_extra'] = exc_extra(exc)
    elif response:
        extra.update(
            response_status=response.status_code,
            response_headers=dict(response.headers),
            response_body=lenient_json(await get_response_body(response)),
        )

    return dict(
        extra=extra,
        user=dict(ip_address=get_ip(request)),
        transaction=re.sub(r'/\d{2,}/', '/{number}/', str(request.url.path)),
        request=dict(
            url=str(request.url),
            query_string=request.url.query,
            cookies=dict(request.cookies),
            headers=dict(request.headers),
            method=request.method,
            data=lenient_json(request.scope.get('_body')),
            inferred_content_type=request.headers.get('Content-Type'),
        ),
    )


async def get_response_body(response: Response) -> bytes:
    if hasattr(response, 'body'):
        return response.body
    else:
        body_chunks = []
        async for chunk in response.body_iterator:
            if not isinstance(chunk, bytes):
                chunk = chunk.encode(response.charset)
            body_chunks.append(chunk)

        response.body_iterator = async_gen_list(body_chunks)
        return b''.join(body_chunks)


async def async_gen_list(list_: List[bytes]) -> AsyncGenerator[bytes, None]:
    for c in list_:
        yield c


def line_one(request: Request) -> str:
    line = f'{request.method} {request.url.path}'
    if q := request.url.query:
        line += f'?{q}'
    return line


def get_request_start(request):
    try:
        return float(request.headers.get('X-Request-Start', '.')) / 1000
    except ValueError:
        return time()


def exc_extra(exc: Exception):
    exception_extra = getattr(exc, 'extra', None)
    if exception_extra:
        try:
            v = exception_extra()
        except Exception:
            pass
        else:
            return lenient_json(v)


def lenient_json(v: Any) -> Any:
    if isinstance(v, (str, bytes)):
        try:
            return json.loads(v)
        except (ValueError, TypeError):
            pass
    return v


session_id_key = 'session_id'
no_cookie_response = """\
{
  "message": "Permission Denied, no session set, updates not permitted"
}
"""
header_error_response = """\
{
  "message": "Permission Denied, %s"
}
"""


def get_session_id(request: Request) -> str:
    return request.session[session_id_key]


def update_session_id(request: Request) -> str:
    request.session[session_id_key] = new_session_id = secrets.token_urlsafe()
    return new_session_id


class CsrfMiddleware(BaseHTTPMiddleware):
    """
    Ensures a GET request has been made before post requests and that session_id is set in the session.

    This prevents CSRF especially if the cookie has same_site strict
    """

    def __init__(
        self,
        app: Starlette,
        *,
        should_check: Callable[[Request], bool] = None,
        enable_header_check: bool = None,
        allows_origins: Set[str] = None,
    ):
        super().__init__(app)
        self.should_check = should_check
        if enable_header_check is None:
            self.enable_header_check = not (glove.settings.dev_mode or glove.settings.test_mode)
        else:
            self.enable_header_check = enable_header_check
        self.allows_origins = glove.settings.origin if allows_origins is None else allows_origins

    async def dispatch(self, request: Request, call_next: CallNext) -> Response:
        if self.should_check and not self.should_check(request):
            return await call_next(request)

        session_id = request.session.get(session_id_key)
        benign_request = request.method in {'HEAD', 'GET', 'OPTIONS'}
        if not benign_request:
            if session_id is None:
                return Response(no_cookie_response, media_type='application/json', status_code=403)
            if error := self.header_check(request):
                return Response(header_error_response % error, media_type='application/json', status_code=403)

        response = await call_next(request)

        # set the session id for any valid GET request
        if benign_request and response.status_code == 200 and session_id is None:
            update_session_id(request)
        return response

    def header_check(self, request: Request) -> Optional[str]:
        """
        Origin and Referrer checks for CSRF, see
        https://cheatsheetseries.owasp.org/cheatsheets/
        Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#verifying-origin-with-standard-headers
        """
        if not self.enable_header_check:
            return

        # origin header is optional (e.g. is omitted from firefox uploads)
        origin_ok = False
        if origin := request.headers.get('origin'):
            if origin not in self.allows_origins:
                return 'Incorrect Origin header'
            origin_ok = True

        if referrer := request.headers.get('referer'):
            referrer_url = URL(referrer)
            if f'{referrer_url.scheme}://{referrer_url.hostname}' not in self.allows_origins:
                return 'Incorrect Referrer header'
        elif not origin_ok:
            return 'Missing Origin and Referrer headers'


class HostRedirectMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: Starlette, host: str = None):
        super().__init__(app)
        self.host = host or glove.settings.domain
        assert self.host, 'host must not be None in HostRedirectMiddleware'

    async def dispatch(self, request: Request, call_next: CallNext) -> Response:
        if request.url.hostname == self.host:
            return await call_next(request)
        else:
            return RedirectResponse(request.url.replace(hostname=self.host), status_code=301)


class IPRangeCounter:
    __slots__ = 'range', 'counter'

    def __init__(self, network_range: str):
        self.range = ip_network(network_range)
        self.counter = 0

    def __repr__(self):
        return f'IPRangeCounter({self.range}, {self.counter})'


class CloudflareCheckMiddleware(BaseHTTPMiddleware):
    default_response_body = b'Request incorrectly routed, this looks like a problem with your DNS or Proxy.'

    def __init__(self, app: Starlette, response_text: str = None):
        super().__init__(app)
        self.response_body = response_text.encode() if response_text else self.default_response_body
        self.ip_ranges: Optional[List[IPRangeCounter]] = None

    async def dispatch(self, request: Request, call_next: CallNext) -> Response:
        """
        On Heroku (and any properly configured system) we can trust the last entry in X-Forwarded-For,
        hence using that before trying client

        https://stackoverflow.com/a/37061471/949890
        """
        ip = None
        x_forwarded_for = request.headers.get('x-forwarded-for')
        if x_forwarded_for:
            ip = x_forwarded_for.rsplit(',', 1)[-1].strip()
        else:
            client = request.scope.get('client')
            if client:
                ip = client[0]

        if ip and await self.is_cloudflare_ip(ip):
            return await call_next(request)

        extra = {'extra': await request_log_extra(request), 'cf_ip_ranges': self.ip_ranges}
        logger.warning('Request not routed through CloudFlare ip=%s url="%s"', ip, request.url, extra=extra)
        return Response(self.response_body, status_code=400)

    async def is_cloudflare_ip(self, ip: str) -> bool:
        try:
            ip = ip_address(ip.strip())
        except ValueError:
            pass
        else:
            if self.ip_ranges is None:
                self.ip_ranges = ip_ranges = await get_cloudflare_ips()
            else:
                ip_ranges = self.ip_ranges

            for r in ip_ranges:
                if ip in r.range:
                    r.counter += 1
                    self.ip_ranges.sort(key=attrgetter('counter'), reverse=True)
                    return True
        return False


async def get_cloudflare_ips() -> List[IPRangeCounter]:
    """
    Get a list of cloudflare IPs from https://www.cloudflare.com/ips-v4 and https://www.cloudflare.com/ips-v6,
    see https://www.cloudflare.com/en-gb/ips/ for details.
    """

    async def get_ips(v: Literal[4, 6]) -> List[IPRangeCounter]:
        r = await glove.http.get(f'https://www.cloudflare.com/ips-v{v}')
        UnexpectedResponse.check(r)
        return [IPRangeCounter(ip) for ip in r.text.strip().split('\n')]

    v4_ips, v6_ips = await asyncio.gather(get_ips(4), get_ips(6))
    ips = v4_ips + v6_ips
    logger.info('downloaded %d IPs from CloudFlare to check requests against', len(ips))
    return ips



from typing import Dict, List, Optional, TypeVar

from starlette.requests import Request

__all__ = 'get_ip', 'list_not_none', 'dict_not_none'

IP_HEADER = 'X-Forwarded-For'


def get_ip(request: Request) -> Optional[str]:
    ips = request.headers.get(IP_HEADER)
    if ips:
        return ips.split(',', 1)[0].strip(' ')
    elif client := request.scope.get('client'):
        return client[0]


T = TypeVar('T')


def list_not_none(*items: T) -> List[T]:
    return [item for item in items if item is not None]


def dict_not_none(*args: Dict[str, T], **kwargs: T) -> Dict[str, T]:
    d = {}
    if args:
        if len(args) > 1:
            raise TypeError(f'dict_not_none expected at most 1 argument, got {len(args)}')
        d = args[0]
        if not isinstance(d, dict):
            raise TypeError(f'dict_not_none must be a dict, got {d.__class__.__name__}')
    if kwargs:
        d.update(kwargs)
    return {key: value for key, value in d.items() if value is not None}



import asyncio
import logging
from typing import Optional

from async_timeout import timeout
from asyncpg import PostgresError
from buildpg.asyncpg import BuildPgConnection, connect_b

from ..settings import BaseSettings

__all__ = 'AsyncPgContext', 'lenient_conn'

logger = logging.getLogger('foxglove.db')


class AsyncPgContext:
    def __init__(self, pg_dsn: str):
        self._pg_dsn = pg_dsn
        self._conn: Optional[BuildPgConnection] = None

    async def __aenter__(self) -> BuildPgConnection:
        self._conn = await connect_b(dsn=self._pg_dsn)
        return self._conn

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._conn:
            await self._conn.close()


async def lenient_conn(settings: BaseSettings, *, with_db: bool = True, sleep: float = 1) -> BuildPgConnection:
    if with_db:
        dsn = settings.pg_dsn
    else:
        dsn, _ = settings.pg_dsn.rsplit('/', 1)

    for retry in range(8, -1, -1):
        try:
            async with timeout(2):
                conn = await connect_b(dsn=dsn)
        except (PostgresError, OSError) as e:
            if retry == 0:
                raise
            else:
                logger.warning('pg temporary connection error "%s", %d retries remaining...', e, retry)
                await asyncio.sleep(sleep)
        else:
            log = logger.debug if retry == 8 else logger.info
            log('pg connection successful, version: %s', await conn.fetchval('SELECT version()'))
            return conn



import logging
import logging.config
import os
import traceback
from copy import copy
from functools import lru_cache
from io import StringIO
from typing import Any, Dict

from uvicorn.logging import DefaultFormatter

from .main import glove

try:
    import pygments
    from pygments.formatters import Terminal256Formatter
    from pygments.lexers import Python3TracebackLexer
except ImportError:  # pragma: no cover
    pyg_lexer = pyg_formatter = None
else:
    pyg_lexer, pyg_formatter = Python3TracebackLexer(), Terminal256Formatter(style='vim')

try:
    from devtools import pformat as format_extra
    from devtools.ansi import isatty, sformat
except ImportError:  # pragma: no cover
    from pprint import pformat

    isatty = False
    sformat = None

    def format_extra(extra, highlight):
        return pformat(extra)


logger = logging.getLogger('foxglove.logs')

# only way to get "extra" from a LogRecord is to look in record.__dict__ and ignore all the standard keys
standard_record_keys = {
    'name',
    'msg',
    'args',
    'levelname',
    'levelno',
    'pathname',
    'filename',
    'module',
    'exc_info',
    'exc_text',
    'stack_info',
    'lineno',
    'funcName',
    'created',
    'msecs',
    'relativeCreated',
    'thread',
    'threadName',
    'processName',
    'process',
    'message',
    'color_message',
}


class HighlightExtraFormatter(DefaultFormatter):
    def __init__(self, *args, **kwargs):
        self.sentry_active = bool(glove.settings.sentry_dsn)
        super().__init__(*args, **kwargs)

    def formatMessage(self, record):
        record_copy = copy(record)
        record_copy.__dict__['nameprefix'] = f'{record_copy.name:28}'
        s = super().formatMessage(record_copy)
        if not self.sentry_active:
            extra = {k: v for k, v in record.__dict__.items() if k not in standard_record_keys}
            if extra:
                s += '\nExtra: ' + format_extra(extra, highlight=self.should_use_colors())
        return s

    def formatException(self, ei):
        if self.sentry_active:
            return super().formatException(ei)
        sio = StringIO()
        traceback.print_exception(*ei, file=sio)
        stack = sio.getvalue()
        sio.close()
        if self.should_use_colors() and pyg_lexer:
            return pygments.highlight(stack, lexer=pyg_lexer, formatter=pyg_formatter).rstrip('\n')
        else:
            return stack


class NotWarnings(logging.Filter):
    def filter(self, record):
        return record.levelno < logging.WARNING


def get_env_multiple(*names):
    for name in names:
        v = os.getenv(name, None) or os.getenv(name.lower(), None)
        if v:
            return v


@lru_cache
def setup_sentry() -> bool:
    if glove.settings.sentry_dsn:
        import sentry_sdk
        from sentry_sdk.integrations.logging import LoggingIntegration, ignore_logger

        sentry_sdk.init(
            dsn=glove.settings.sentry_dsn,
            integrations=[LoggingIntegration(level=logging.INFO, event_level=logging.WARNING)],
            release=glove.settings.release,
            environment=glove.settings.environment,
            server_name=get_env_multiple('DYNO', 'SERVER_NAME', 'HOSTNAME', 'HOST', 'NAME'),
        )
        ignore_logger('foxglove.bad_requests')
        logger.info('sentry initialised')
        return True
    else:
        logger.info('sentry not initialised')
        return False


def build_logging_config() -> Dict[str, Any]:
    """
    setup logging config by updating the arq logging config
    """
    log_level = glove.settings.log_level
    config = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'foxglove.default': {
                '()': 'foxglove.logs.HighlightExtraFormatter',
                'fmt': '%(nameprefix)s%(levelprefix)s %(message)s',
                'use_colors': None,
            },
            'foxglove.access': {
                '()': 'uvicorn.logging.AccessFormatter',
                'fmt': "%(levelprefix)s %(client_addr)s - '%(request_line)s' %(status_code)s",
            },
        },
        'handlers': {
            'foxglove.default': {
                'formatter': 'foxglove.default',
                'class': 'logging.StreamHandler',
                'stream': 'ext://sys.stderr',
                'level': log_level,
            },
            'foxglove.access': {
                'formatter': 'foxglove.access',
                'class': 'logging.StreamHandler',
                'stream': 'ext://sys.stdout',
                'level': log_level,
            },
        },
        'loggers': {
            '': {'handlers': ['foxglove.default'], 'level': log_level},
            'uvicorn.error': {'handlers': ['foxglove.default'], 'level': log_level, 'propagate': False},
            'uvicorn.access': {'handlers': ['foxglove.access'], 'level': log_level, 'propagate': False},
        },
    }
    return config


def setup_logging() -> None:
    config = build_logging_config()
    logging.config.dictConfig(config)
    setup_sentry()
